@@ -60,9 +60,10 @@ void Response::Cookie::serialize_to (string& acc, const string& name, const Requ
     }
 }
 
-string Response::_http_header (const Request* req, Compression::Type applied_compression) const {
+string Response::_http_header (SerializationContext &ctx) const {
     //part 1: precalc pieces
-    auto tmp_http_ver = this->http_version;
+    auto req = ctx.request;
+    auto tmp_http_ver = ctx.http_version;
     auto tmp_code = code ? code : 200;
 
     auto out_connection = headers.get("Connection");
@@ -82,19 +83,21 @@ string Response::_http_header (const Request* req, Compression::Type applied_com
 
     string out_content_length;
     if (!chunked && !headers.has("Content-Length")) {
-        out_content_length = panda::to_string(body.length());
+        out_content_length = panda::to_string(ctx.body->length());
     }
 
-    auto out_content_encoding = _content_encoding(applied_compression);
+    auto out_content_encoding = _content_encoding(ctx);
 
     // part 2: summarize pieces size
     size_t reserved = 5 + 4 + 4 + out_mesasge.length() + 2 + headers.length() + 2;
-    if (out_connection)       reserved += 10 + 2 + out_connection.length()       + 2;
+    if (out_connection)       reserved += 10 + 2 + out_connection.length()   + 2;
     if (out_content_length)   reserved += 14 + 2 + out_content_length.length()   + 2;
     if (out_content_encoding) reserved += 16 + 2 + out_content_encoding.length() + 2;
 
+    for (auto& h: ctx.handled_headers) reserved += h.name.length() + 2 + h.value.length() + 2;
     for (auto& h: headers) {
-        if (h.name == "Connection") continue; // already handled
+        if (ctx.handled_headers.has(h.name)) continue;  // let's speedup a little bit
+        if (h.name == "Connection") continue;  // already handled
         reserved += h.name.length() + 2 + h.value.length() + 2;
     };
     reserved += (200 + 14) * cookies.fields.size(); // should be enough for average set-cookie header
@@ -118,8 +121,10 @@ string Response::_http_header (const Request* req, Compression::Type applied_com
     if (out_content_length)   { s += "Content-Length: "  ; s += out_content_length  ; s += "\r\n" ;};
     if (out_content_encoding) { s += "Content-Encoding: "; s += out_content_encoding; s += "\r\n" ;};
 
+    for (auto& h: ctx.handled_headers) { s += h.name; s += ": "; s += h.value; s+= "\r\n"; }
     for (auto& h: headers)  {
-        if (h.name == "Connection") continue; // already handled
+        if (ctx.handled_headers.has(h.name)) continue;
+        if (h.name == "Connection") continue;  // already handled
         s += h.name; s += ": "; s += h.value; s+= "\r\n";
     };
     for (auto& c: cookies.fields) {
@@ -134,14 +139,18 @@ string Response::_http_header (const Request* req, Compression::Type applied_com
     return s;
 }
 
-std::vector<string> Response::to_vector (const Request* req) {
+std::vector<string> Response::to_vector (const Request* req) const {
     /* if client didn't announce Accept-Encoding or we do not support it, just pass data as it is */
     auto applied_compression
             = req && (compression.type != Compression::IDENTITY) && (req->allowed_compression() & (uint8_t)compression.type)
             ? compression.type
             : Compression::IDENTITY;
 
-    return _to_vector(applied_compression, [&]{ return _http_header(req,  applied_compression); });
+    SerializationContext ctx;
+    ctx.compression = applied_compression;
+    ctx.request = req;
+    ctx.body = &body;
+    return _to_vector(ctx, [&]() { return _compile_prepare(ctx); }, [&]() { return _http_header(ctx); });
 }
 
 string Response::message_for_code (int code) {
